@@ -4,7 +4,7 @@ use 5.006;
 use strict;
 use warnings;
 
-our $VERSION = '0.21';
+our $VERSION = '0.24';
 
 # list compatible repository versions
 # This is a list of numbers of the form "\d+.\d".
@@ -14,22 +14,25 @@ our $VERSION = '0.21';
 # first digit after the period.
 our $Compatible_Versions = {
     '0.1' => 1,
+    '0.2' => 1,
 };
 
-use constant MODULES_DBM_FILE     => 'modules_dists.dbm';
-use constant SYMLINKS_DBM_FILE    => 'symlinks.dbm';
-use constant SCRIPTS_DBM_FILE     => 'scripts_dists.dbm';
-use constant REPOSITORY_INFO_FILE => 'repository_info.yml';
-use constant DBM_CHECKSUMS_FILE   => 'dbm_checksums.txt';
+use constant MODULES_DBM_FILE      => 'modules_dists.dbm';
+use constant SYMLINKS_DBM_FILE     => 'symlinks.dbm';
+use constant SCRIPTS_DBM_FILE      => 'scripts_dists.dbm';
+use constant DEPENDENCIES_DBM_FILE => 'dependencies.dbm';
+use constant REPOSITORY_INFO_FILE  => 'repository_info.yml';
+use constant DBM_CHECKSUMS_FILE    => 'dbm_checksums.txt';
 
-use base 'PAR::Repository::Query';
-use base 'PAR::Repository::Client::Util';
+use base 'PAR::Repository::Query',
+         'PAR::Repository::Client::Util',
+         'PAR::Repository::Client::DBM';
 
 require PAR::Repository::Client::HTTP;
 require PAR::Repository::Client::Local;
 
 use Carp qw/croak/;
-use File::Spec::Functions qw/splitpath/;
+use File::Spec;
 require version;
 require Config;
 require PAR::Dist;
@@ -37,6 +40,7 @@ require DBM::Deep;
 require Archive::Zip;
 require File::Temp;
 require File::Copy;
+require File::Path;
 require YAML::Tiny;
 
 =head1 NAME
@@ -66,14 +70,14 @@ PAR::Repository::Client - Access PAR repositories
 This module represents the client for PAR repositories as
 implemented by the L<PAR::Repository> module.
 
-Chances are, you should be looking at the L<PAR> module
+Chances are you should be looking at the L<PAR> module
 instead. Starting with version 0.950, it supports
 automatically loading any modules that aren't found on your
 system from a repository. If you need finer control than that,
 then this module is the right one to use.
 
 You can use this module to access repositories in one of
-two ways: On your local filesystem or via HTTP. The
+two ways: On your local filesystem or via HTTP(S?). The
 access methods are implemented in
 L<PAR::Repository::Client::HTTP> and L<PAR::Repository::Client::Local>.
 Any common code is in this module.
@@ -126,6 +130,14 @@ specified module does not exist on the local system yet or is outdated.
 You cannot set both I<auto_install> and I<auto_upgrade>. If you do,
 you will get a fatal error.
 
+If you set the C<static_dependencies> option to a true value,
+then the inter-distribution dependency information that is retrieved
+from the repository will be used to recursively apply your requested
+action to all dependencies. Essentially, this makes the C<install_module>
+method act like a real package manager similar to PPM.
+In contrast, the default behaviour is to fetch distributions only
+on demand and potentially recursively.
+
 In order to control where the modules are installed to, you can
 use the C<installation_targets> method.
 
@@ -141,6 +153,14 @@ compatible with this version of the client.
 
 You may specify a C<http_timeout> in seconds.
 
+The C<cache_dir>
+parameter can be used to set the directory where you want the downloaded
+files to reside. It defaults to the C<$ENV{PAR_TEMP}> directory or
+otherwise the C<par> subdirectory of your system's temporary directory.
+If you set C<cache_dir> to something other than the default, the downloaded
+files should be automatically cached when the HTTP transport layer is
+used as C<LWP::mirror()> only checks for updates.
+
 =cut
 
 sub new {
@@ -167,28 +187,44 @@ sub new {
 
   my $self = bless {
     # the repository uri
-    uri => $uri,
+    uri                   => $uri,
+
     # The last error message
-    error => '',
+    error                 => '',
+
     # The hash ref of checksums for checking whether we
     # need to update the dbms
-    checksums => undef,
-    supports_checksums => undef,
+    checksums             => undef,
+    supports_checksums    => undef,
+
     # the modules- and scripts dbm storage
     # both the local temp file for cleanup
     # and the actual tied hash
-    modules_dbm_temp_file => undef,
-    modules_dbm_hash => undef,
-    scripts_dbm_temp_file => undef,
-    scripts_dbm_hash => undef,
-    info => undef, # used for YAML info caching
-    auto_install => $args{auto_install},
-    auto_upgrade => $args{auto_upgrade},
-    installation_targets => {}, # see PAR::Dist
-    perl_version => (defined($args{perl_version}) ? $args{perl_version} : $Config::Config{version}),
-    architecture => (defined($args{architecture}) ? $args{architecture} : $Config::Config{archname}),
-  } => "PAR::Repository::Client::$obj_class";
+    modules_dbm_temp_file      => undef,
+    modules_dbm_hash           => undef,
+    scripts_dbm_temp_file      => undef,
+    scripts_dbm_hash           => undef,
+    dependencies_dbm_temp_file => undef,
+    dependencies_dbm_hash      => undef,
 
+    info                  => undef, # used for YAML info caching
+    auto_install          => $args{auto_install},
+    auto_upgrade          => $args{auto_upgrade},
+    static_dependencies   => $args{static_dependencies},
+    installation_targets  => {}, # see PAR::Dist
+    perl_version          => (defined($args{perl_version}) ? $args{perl_version} : $Config::Config{version}),
+    architecture          => (defined($args{architecture}) ? $args{architecture} : $Config::Config{archname}),
+    cache_dir             => $args{cache_dir},
+  } => "PAR::Repository::Client::$obj_class";
+  
+  if (not defined $self->{cache_dir}) {
+    $self->{cache_dir} = defined($ENV{PAR_TEMP})
+                         ? $ENV{PAR_TEMP}
+                         : File::Spec->catdir(File::Spec->tmpdir, 'par');
+  }
+  
+  File::Path::mkpath($self->{cache_dir});
+  
   $self->_init(\%args);
 
   $self->validate_repository()
@@ -258,7 +294,7 @@ sub use_module {
   my $required = $self->require_module($namespace);
   return() if not $required; # error set by require_module
 
-    eval "package $pkg; ${namespace}->import(\@args) if ${namespace}->can('import');";
+  eval "package $pkg; ${namespace}->import(\@args) if ${namespace}->can('import');";
   if ($@) {
     $self->{error} = "An error occurred while executing 'package $pkg; ${namespace}->import(\@args);'. Error: $@";
     return();
@@ -276,7 +312,9 @@ loading preferably.)
 Searches for a specified namespace in the repository and downloads
 the corresponding PAR distribution. Automatically loads PAR
 and appends the downloaded PAR distribution to the list of
-PARs to load from.
+PARs to load from. If auto-installation or auto-upgrading was
+enabled, the contents of the PAR distribution will
+be installed in addition to loading the PAR.
 
 Returns the name of the local
 PAR file. Think of this as C<require_module> without actually
@@ -292,55 +330,40 @@ sub get_module {
 
   $self->{error} = undef;
 
-  my $local_par_file;
+  my @local_par_files;
   if ($self->{auto_install}) {
-    $local_par_file = $self->install_module($namespace);
+    @local_par_files = $self->install_module($namespace);
   }
   elsif ($self->{auto_upgrade}) {
-    $local_par_file = $self->upgrade_module($namespace);
+    @local_par_files = $self->upgrade_module($namespace);
+  }
+  elsif ($self->{static_dependencies}) {
+    my $deps = $self->get_module_dependencies($namespace);
+    return() if not ref $deps;
+
+    foreach my $dep_dist (@$deps) {
+      my $local_par_file = $self->_fetch_dist($dep_dist);
+      return() if not defined $local_par_file;
+      push @local_par_files, $local_par_file;
+    }
   }
   else {
-    $local_par_file = $self->_fetch_module($namespace);
+    my $dist = $self->_module2dist($namespace);
+    return() if not defined $dist;
+    my $local_par_file = $self->_fetch_dist($dist);
     return() if not defined $local_par_file;
+    push @local_par_files, $local_par_file;
+  }
+  return() if not @local_par_files;
 
-    require PAR;
+  require PAR;
+  foreach my $local_par_file ($fallback ? @local_par_files : reverse(@local_par_files)) {
     PAR->import( { file => $local_par_file, fallback => ($fallback?1:0) } );
   }
-  return() if not defined $local_par_file;
 
-  return $local_par_file;
+  return shift @local_par_files; # FIXME should this return the full array?
 }
 
-sub _fetch_module {
-  my $self = shift;
-  my $namespace = shift;
-
-  $self->{error} = undef;
-
-  my ($modh) = $self->modules_dbm;
-  if (not defined $modh) {
-    return();
-  }
-
-  my $dists = $modh->{$namespace};
-  if (not defined $dists) {
-    $self->{error} = "Could not find module '$namespace' in the repository.";
-    return();
-  }
-
-  my $dist = $self->prefered_distribution($namespace, $dists);
-  if (not defined $dist) {
-    $self->{error} = "PAR: Could not find a distribution for package '$namespace'";
-    return();
-  }
-
-  my $local_par_file = $self->fetch_par($dist);
-  if (not defined $local_par_file or not -f $local_par_file) {
-    return();
-  }
-
-  return $local_par_file;
-}
 
 =head2 install_module
 
@@ -366,15 +389,30 @@ sub install_module {
 
   $self->{error} = undef;
 
-  my $local_par_file = $self->_fetch_module($namespace);
-  return() if not defined $local_par_file;
+  my @local_par_files;
+  if ($self->{static_dependencies}) {
+    my $deps = $self->get_module_dependencies($namespace);
+    return() if not ref $deps;
 
-  PAR::Dist::install_par(
-    %{$self->installation_targets()},
-    dist => $local_par_file,
-  ) or return ();
+    foreach my $dep_dist (@$deps) {
+      my $local_par_file = $self->_fetch_dist($dep_dist);
+      return() if not defined $local_par_file;
+      push @local_par_files, $local_par_file;
+    }
+  }
+  else {
+    push @local_par_files, $self->_fetch_module($namespace);
+  }
+  return() if not @local_par_files;
 
-  return $local_par_file;
+  foreach my $local_par_file (@local_par_files) {
+    PAR::Dist::install_par(
+      %{$self->installation_targets()},
+      dist => $local_par_file,
+    ) or return ();
+  }
+
+  return shift @local_par_files; # FIXME should this return the whole array?
 }
 
 
@@ -397,6 +435,17 @@ Returns the name of the local F<.par> file which was installed or
 the empty list on failure or if the local version of the module is
 already current.
 
+I<CAVEAT:> This will first try to require a locally installed version
+of the module. If that succeeds, its version is compared to the
+highest version in the repository. If an upgrade is necessary,
+the new module will be installed. If the module hadn't been found
+locally before the installation, it will be loaded. If it was
+found locally (and thus loaded), C<IT WILL NOT BE RELOADED SO
+YOU GET THE NEW VERSION>.
+This is because reloading of modules is not a simple issue.
+If you need this behaviour, you can get it manually using L<Class::Unload>
+and another require.
+
 =cut
 
 sub upgrade_module {
@@ -407,14 +456,13 @@ sub upgrade_module {
 
   # get local version
   my $local_version;
-  eval "require $namespace; \$local_version = ${namespace}->VERSION;";
+  local @PAR::RepositoryObjects = (); # do not load from remote!
+  local @PAR::UpgradeRepositoryObjects = ();
+  eval "require ${namespace}; \$local_version = ${namespace}->VERSION;";
+  $local_version = version->new($local_version) if defined($local_version) and not eval {$local_version->isa('version')};
 
   # no local version found. Install from repo
   if (not defined $local_version) {
-    my $mod = $namespace;
-    $mod =~ s/::/\//;
-    $mod .= '.pm';
-    delete $INC{$mod};
     return $self->install_module($namespace);
   }
 
@@ -438,12 +486,9 @@ sub upgrade_module {
   }
 
   my $repo_version = $modh->{$namespace}{$dist};
+  $repo_version = version->new($repo_version) if not eval {$repo_version->isa('version')};
 
   if ($repo_version > $local_version) {
-    my $mod = $namespace;
-    $mod =~ s/::/\//;
-    $mod .= '.pm';
-    delete $INC{$mod};
     return $self->install_module($namespace);
   }
 
@@ -469,56 +514,117 @@ scripts. This differs from the behaviour for mdoules.
 
 =cut
 
-
 sub run_script {
   my $self = shift;
   my $script = shift;
 
-  $self->{error} = undef;
+  my @local_par_files;
+  if ($self->{static_dependencies}) {
+    my $deps = $self->get_script_dependencies($script);
+    return() if not ref $deps;
 
-  my ($scrh) = $self->scripts_dbm;
-  if (not defined $scrh) {
-    return();
+    foreach my $dep_dist (@$deps) {
+      my $local_par_file = $self->_fetch_dist($dep_dist);
+      return() if not defined $local_par_file;
+      push @local_par_files, $local_par_file;
+    }
   }
-
-  my $dists = $scrh->{$script};
-  if (not defined $dists) {
-    $self->{error} = "Could not find script '$script' in the repository.";
-    return();
-  }
-  my $dist = $self->prefered_distribution($script, $dists);
-  if (not defined $dist) {
-    $self->{error} = "PAR: Could not find a distribution for script '$script'";
-    return();
-  }
-
-  my $local_par_file = $self->fetch_par($dist);
-  if (not defined $local_par_file or not -f $local_par_file) {
-    return();
+  else {
+    my $dist = $self->_script2dist($script);
+    return() unless defined $dist;
+    my $local_par_file = $self->fetch_par($dist);
+    return() unless defined $local_par_file;
+    push @local_par_files, $local_par_file;
   }
+  return() if not @local_par_files;
 
   if ($self->{auto_install}) {
-    PAR::Dist::install_par(
+    foreach my $local_par_file (@local_par_files) {
+      PAR::Dist::install_par(
         %{ $self->installation_targets() },
         dist => $local_par_file,
-        ) or return ();
+      ) or return ();
+    }
   }
   elsif ($self->{auto_upgrade}) {
     # FIXME This is not the right way to do it!
-    PAR::Dist::install_par(
+    foreach my $local_par_file (@local_par_files) {
+      PAR::Dist::install_par(
         %{ $self->installation_targets() },
         dist => $local_par_file,
-        ) or return ();
+      ) or return ();
+    }
   }
-  
+
   require PAR;
-  PAR->import( { file => $local_par_file, run => $script } );
+  my $script_par = shift @local_par_files;
+  foreach my $local_par_file (@local_par_files) {
+    PAR->import( { file => $local_par_file } );
+  }
+
+  PAR->import( { file => $script_par, run => $script } );
 
   # doesn't happen!?
   return 1;
 }
 
 
+=head2 get_module_dependencies
+
+Given a module name, determines the correct distribution in
+the repository that supplies the module. Returns a reference
+to an array containing that distribution and all distributions
+it depends on. The distribution that contains the given module
+is the first in the array.
+
+Returns the empty list on failure.
+
+=cut
+
+sub get_module_dependencies {
+  my $self = shift;
+  my $namespace = shift;
+  $self->{error} = undef,
+
+  my $dist = $self->_module2dist($namespace);
+  return() if not defined $dist;
+
+  my $deps = $self->_resolve_static_dependencies($dist);
+  return() if not ref $deps;
+  unshift @$deps, $dist;
+
+  return $deps;
+}
+
+
+=head2 get_script_dependencies
+
+Given a script name, determines the correct distribution in
+the repository that supplies the script. Returns a reference
+to an array containing that distribution and all distributions
+it depends on. The distribution that contains the given script
+is the first in the array.
+
+Returns the empty list on failure.
+
+=cut
+
+sub get_script_dependencies {
+  my $self = shift;
+  my $script = shift;
+  $self->{error} = undef,
+
+  my $dist = $self->_script2dist($script);
+  return() if not defined $dist;
+
+  my $deps = $self->_resolve_static_dependencies($dist);
+  return() if not ref $deps;
+  unshift @$deps, $dist;
+
+  return $deps;
+}
+
+
 =head2 installation_targets
 
 Sets the installation targets for modules and scripts if any arguments are
@@ -550,6 +656,10 @@ sub installation_targets {
 }
 
 
+=head1 ACCESSORS
+
+These methods get or set some attributes of the repository client.
+
 =head2 error
 
 Returns the last error message if there was an error or
@@ -557,16 +667,12 @@ the empty list otherwise.
 
 =cut
 
-
 sub error {
   my $self = shift;
   my $err = $self->{error};
   return(defined($err) ? $err : ());
 }
 
-=head1 ACCESSORS
-
-These methods get or set some attributes of the repository client.
 
 =head2 perl_version
 
@@ -588,6 +694,7 @@ sub perl_version {
   return $self->{perl_version};
 }
 
+
 =head2 architecture 
 
 Sets and/or returns the name of the architecture which is used to choose the right
@@ -608,6 +715,7 @@ sub architecture {
   return $self->{architecture};
 }
 
+
 =head1 OTHER METHODS
 
 These methods, while part of the official interface, should need rarely be
@@ -681,6 +789,7 @@ sub prefered_distribution {
   return $dist->[0];
 }
 
+
 =head2 validate_repository_version
 
 Accesses the repository meta information and validates that it
@@ -714,312 +823,113 @@ sub validate_repository_version {
     $self->{error} = "Repository has an incompatible version (".$info->[0]{repository_version}.")";
     return();
   }
-  return 1;
-}
 
-=head2 need_dbm_update
-
-Takes one or no arguments. Without arguments, all DBM files are
-checked. With an argument, only the specified DBM file will be checked.
-
-Returns true if either one of the following conditions match:
-
-=over 2
-
-=item
-
-The repository does not support checksums.
-
-=item
-
-The checksums (and thus also the DBM files) haven't been
-downloaded yet.
-
-=item
-
-The local copies of the checksums do not match those of the repository.
-
-=back
-
-In cases two and three above, the return value is actually the hash
-reference of checksums that was fetched from the repository.
-
-Returns the empty list if the local checksums match those of the
-repository exactly.
-
-You don't usually need to call this directly. By default, DBM files
-are only fetched from the repository if necessary.
-
-=cut
-
-sub need_dbm_update {
-  my $self = shift;
-  $self->{error} = undef;
-
-  my $check_file = shift;
-  $check_file .= '.zip' if defined $check_file and not $check_file =~ /\.zip$/;
-
-  my $support = $self->{supports_checksums};
-  if (defined $support and not $support) {
-    return 1;
-  }
-
-  my $checksums = $self->_dbm_checksums();
-
-  if (not defined $checksums) {
-    $self->{supports_checksums} = 0;
-    return 1;
-  }
-  else {
-    $self->{supports_checksums} = 1;
-  }
-
-  if (not defined $self->{checksums}) {
-    # never fetched checksums before.
-    return $checksums;
-  }
-  else {
-    # we fetched checksums earlier, match them
-    my $local_checksums = $self->{checksums};
-    if (not defined $check_file) {
-      return $checksums if keys(%$local_checksums) != keys(%$checksums);
-      foreach my $file (keys %$checksums) {
-        return $checksums
-          if not exists $local_checksums->{$file}
-          or not $local_checksums->{$file} eq $checksums->{$file};
-      }
-    }
-    else {
-      return $checksums
-        if not exists $local_checksums->{$check_file}
-        or not exists $checksums->{$check_file}
-        or not $local_checksums->{$check_file} eq $checksums->{$check_file};
-    }
-    return 0;
-  }
-}
-
-
-=head2 modules_dbm
-
-Fetches the C<modules_dists.dbm> database from the repository,
-ties it to a L<DBM::Deep> object and returns a tied hash
-reference or the empty list on failure. Second return
-value is the name of the local temporary file.
-
-In case of failure, an error message is available via
-the C<error()> method.
-
-The method uses the C<_fetch_dbm_file()> method which must be
-implemented in a subclass such as L<PAR::Repository::Client::HTTP>.
-
-=cut
-
-sub modules_dbm {
-  my $self = shift;
-  $self->{error} = undef;
-
-  my $checksums = $self->need_dbm_update();
-  if ($self->{modules_dbm_hash}) {
-    # need new dbm file?
-    return $self->{modules_dbm_hash} if not $checksums;
-
-    # does this particular dbm need to be updated?
-    if ($self->{checksums}) {
-      my $local_checksum = $self->{checksums}{MODULES_DBM_FILE().".zip"};
-      my $remote_checksum = $checksums->{MODULES_DBM_FILE().".zip"};
-      return $self->{modules_dbm_hash}
-        if defined $local_checksum and defined $remote_checksum
-           and $local_checksum eq $remote_checksum;
-    }
-
-    # just to make sure
-    $self->close_modules_dbm;
-  }
-
-  my $file = $self->_fetch_dbm_file(MODULES_DBM_FILE().".zip");
-  # (error set by _fetch_dbm_file)
-  return() if not defined $file; # or not -f $file; # <--- _fetch_dbm_file should do the stat!
-
-  my ($tempfh, $tempfile) = File::Temp::tempfile(
-    'temporary_dbm_XXXXX',
-    UNLINK => 0,
-    DIR => File::Spec->tmpdir(),
-  );
-
-  if (not $self->_unzip_file($file, $tempfile, MODULES_DBM_FILE())) {
-    $self->{error} = "Could not unzip dbm file '$file' to '$tempfile'";
+  $repo_version =~ s/_.*$//; # remove dev suffix
+  if ($repo_version < 0.18 and $self->{static_dependencies}) {
+    $self->{error} = "Client has static dependency resolution enabled, but repository does not support that. "
+                    ."Either upgrade your repository to version 0.18 or greater or disable static dependency "
+                    ."resolution in the client.";
     return();
   }
 
-  unlink $file;
-
-  $self->{modules_dbm_temp_file} = $tempfile;
-
-  my %hash;
-  my $obj = tie %hash, "DBM::Deep", {
-    file => $tempfile,
-    locking => 1,
-    autoflush => 0,
-  }; 
-
-  $self->{modules_dbm_hash} = \%hash;
-
-  # save this dbm file checksum
-  if (ref($checksums)) {
-    if (not $self->{checksums}) {
-      $self->{checksums} = {};
-    }
-    $self->{checksums}{MODULES_DBM_FILE().".zip"} = $checksums->{MODULES_DBM_FILE().".zip"};
-  }
-
-  return (\%hash, $tempfile);
+  return 1;
 }
 
-=head2 scripts_dbm
 
-Fetches the C<scripts_dists.dbm> database from the repository,
-ties it to a L<DBM::Deep> object and returns a tied hash
-reference or the empty list on failure. Second return
-value is the name of the local temporary file.
-
-In case of failure, an error message is available via
-the C<error()> method.
-
-The method uses the C<_fetch_dbm_file()> method which must be
-implemented in a subclass such as L<PAR::Repository::Client::HTTP>.
-
-=cut
-
-sub scripts_dbm {
+# given a module name, find the prefered distribution
+sub _module2dist {
   my $self = shift;
+  my $namespace = shift;
+
   $self->{error} = undef;
 
-  my $checksums = $self->need_dbm_update();
-  if ($self->{scripts_dbm_hash}) {
-    # need new dbm file?
-    return $self->{scripts_dbm_hash} if not $checksums;
-
-    # does this particular dbm need to be updated?
-    if ($self->{checksums}) {
-      my $local_checksum = $self->{checksums}{SCRIPTS_DBM_FILE().".zip"};
-      my $remote_checksum = $checksums->{SCRIPTS_DBM_FILE().".zip"};
-      return $self->{scripts_dbm_hash}
-        if defined $local_checksum and defined $remote_checksum
-           and $local_checksum eq $remote_checksum;
-    }
-
-    # just to make sure
-    $self->close_scripts_dbm;
-  }
-
-  my $file = $self->_fetch_dbm_file(SCRIPTS_DBM_FILE().".zip");
-  # (error set by _fetch_dbm_file)
-  return() if not defined $file; # or not -f $file; # <--- _fetch_dbm_file should do the stat!
-
-  my ($tempfh, $tempfile) = File::Temp::tempfile(
-    'temporary_dbm_XXXXX',
-    UNLINK => 0,
-    DIR => File::Spec->tmpdir(),
-  );
-
-  if (not $self->_unzip_file($file, $tempfile, SCRIPTS_DBM_FILE())) {
-    $self->{error} = "Could not unzip dbm file '$file' to '$tempfile'";
+  my ($modh) = $self->modules_dbm;
+  if (not defined $modh) {
     return();
   }
 
-  unlink $file;
-
-  $self->{scripts_dbm_temp_file} = $tempfile;
-
-  my %hash;
-  my $obj = tie %hash, "DBM::Deep", {
-    file => $tempfile,
-    locking => 1,
-    autoflush => 0,
-  }; 
-
-  $self->{scripts_dbm_hash} = \%hash;
-
-  # save this dbm file checksum
-  if (ref($checksums)) {
-    if (not $self->{checksums}) {
-      $self->{checksums} = {};
-    }
-    $self->{checksums}{SCRIPTS_DBM_FILE().".zip"} = $checksums->{SCRIPTS_DBM_FILE().".zip"};
+  if (not exists $modh->{$namespace} or not defined $modh->{$namespace}) {
+    $self->{error} = "Could not find module '$namespace' in the repository.";
+    return();
   }
 
-  return (\%hash, $tempfile);
+  my $dist = $self->prefered_distribution($namespace, $modh->{$namespace});
+  if (not defined $dist) {
+    $self->{error} = "PAR: Could not find a distribution for package '$namespace'";
+    return();
+  }
+  return $dist;
 }
 
 
-=head2 close_modules_dbm
-
-Closes the C<modules_dists.dbm> file and does all necessary
-cleaning up.
-
-This is called when the object is destroyed.
-
-=cut
-
-sub close_modules_dbm {
+# resolve a script to its prefered distribution
+sub _script2dist {
   my $self = shift;
-  my $hash = $self->{modules_dbm_hash};
-  return if not defined $hash;
+  my $script = shift;
 
-  my $obj = tied($hash);
-  $self->{modules_dbm_hash} = undef;
-  undef $hash;
-  undef $obj;
+  $self->{error} = undef;
 
-  unlink $self->{modules_dbm_temp_file};
-  $self->{modules_dbm_temp_file} = undef;
-  if ($self->{checksums}) {
-    delete $self->{checksums}{MODULES_DBM_FILE().".zip"};
+  my ($scrh) = $self->scripts_dbm;
+  if (not defined $scrh) {
+    return();
   }
 
-  return 1;
+  my $dists = $scrh->{$script};
+  if (not defined $dists) {
+    $self->{error} = "Could not find script '$script' in the repository.";
+    return();
+  }
+  my $dist = $self->prefered_distribution($script, $dists);
+  if (not defined $dist) {
+    $self->{error} = "PAR: Could not find a distribution for script '$script'";
+    return();
+  }
+  
+  return $dist;
 }
 
-=head2 close_scripts_dbm
 
-Closes the C<scripts_dists.dbm> file and does all necessary
-cleaning up.
-
-This is called when the object is destroyed.
-
-=cut
-
-sub close_scripts_dbm {
+# download a distribution
+sub _fetch_dist {
   my $self = shift;
-  my $hash = $self->{scripts_dbm_hash};
-  return if not defined $hash;
+  my $dist = shift;
 
-  my $obj = tied($hash);
-  $self->{scripts_dbm_hash} = undef;
-  undef $hash;
-  undef $obj;
+  my $local_par_file = $self->fetch_par($dist);
+  return() if not defined $local_par_file or not -f $local_par_file;
 
-  unlink $self->{scripts_dbm_temp_file};
-  $self->{scripts_dbm_temp_file} = undef;
-  if ($self->{checksums}) {
-    delete $self->{checksums}{SCRIPTS_DBM_FILE().".zip"};
-  }
-
-  return 1;
+  return $local_par_file;
 }
 
-=head1 PRIVATE METHODS
 
-These private methods should not be relied upon from the outside of
-the module. (See also: L<PAR::Repository::Client::Util>)
+# resolve a namespace to a distribution and download it
+sub _fetch_module {
+  my $self = shift;
+  my $namespace = shift;
+
+  my $dist = $self->_module2dist($namespace);
+  return() unless defined $dist;
+
+  return $self->_fetch_dist($dist);
+}
+
+
+# resolve a script to a distribution and download it
+sub _fetch_script {
+  my $self = shift;
+  my $namespace = shift;
+
+  my $dist = $self->_script2dist($namespace);
+  return() unless defined $dist;
+
+  return $self->_fetch_dist($dist);
+}
 
-=cut
 
 sub DESTROY {
   my $self = shift;
   $self->close_modules_dbm;
   $self->close_scripts_dbm;
+  $self->close_dependencies_dbm;
 }
 
 1;
@@ -1047,7 +957,7 @@ Steffen Mueller, E<lt>smueller@cpan.orgE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright 2006-2008 by Steffen Mueller
+Copyright 2006-2009 by Steffen Mueller
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.6 or,
